# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from frugal.protocol import FProtocol


class FProtocolFactory(object):
    """FProtocolFactory creates FProtocols."""

    def __init__(self, t_protocol_factory):
        """Initialize FProtocolFactory.

        Args:
            t_protocol_factory: Thrift TProtocolFactory.
        """
        self._t_protocol_factory = t_protocol_factory

    def get_protocol(self, transport):
        """Return FProtocol for the given TTransport.

        Args:
            transport: Thrift TTransport.

        Returns:
            FProtocol wrapping the TTransport.
        """
        return FProtocol(self._t_protocol_factory.getProtocol(transport))
